import os
import ConfigParser

import kivy
kivy.require('1.10.0')
from kivy.uix.button import Button
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.slider import Slider

# sets the location of the config folder
configdir = os.path.join(os.path.expanduser('~'),".kmpc")

class KmpcHelpers(object):

    def formatsong(self,rec):
        """Method used by library browser to properly format a song row."""
        song = ''
        # check if there is more than one disc and display if so
        (d1,d2)=rec['disc'].split('/')
        if int(d2) > 1:
            song+='(Disc '+'%02d' % int(d1)+') '
        # sometimes track numbers are like '01/05' (one of five), so drop that second number
        (t1,t2)=rec['track'].split('/')
        song+='%02d' % int(t1)+' '
        # if albumartist is different than track artist, display the track artist
        if rec['artist'] != rec['albumartist']:
            song+=rec['artist']+' - '
        # display the track title
        song+=rec['title']
        return song

    def loadconfigfile(self):
        # set up config with default values
        config=ConfigParser.SafeConfigParser()
        config.add_section('mpd')
        config.set('mpd','mpdhost','127.0.0.1')
        config.set('mpd','mpdport','6600')
        config.add_section('paths')
        config.set('paths','musicpath','/mnt/music')
        config.set('paths','fanartpath','/mnt/fanart')
        config.set('paths','tmppath','/tmp')
        config.add_section('sync')
        config.set('sync','synchost','127.0.0.1')
        config.set('sync','syncmusicpath','/mnt/music')
        config.set('sync','syncfanartpath','/mnt/fanart')
        config.set('sync','synctmppath','/tmp')
        config.add_section('flags')
        config.set('flags','rpienable','False')
        config.add_section('logs')
        config.set('logs','artlog','False')
        config.add_section('fanart')
        config.set('fanart','client_key','')
        config.add_section('songratings')
        config.set('songratings','zero','Silence')
        config.set('songratings','one','Songs that should never be heard')
        config.set('songratings','two','Songs no one likes')
        config.set('songratings','three','Songs for certain occasions')
        config.set('songratings','four','Songs someone else likes')
        config.set('songratings','five','Filler tracks with no music')
        config.set('songratings','six','Meh track or short musical filler')
        config.set('songratings','seven','Occasional listening songs')
        config.set('songratings','eight','Great songs for all occasions')
        config.set('songratings','nine','Best songs by an artist')
        config.set('songratings','ten','Favorite songs of all time')
        # check if config folder exists
        if os.path.isdir(configdir):
            # try to read existing config file
            config.read([os.path.join(configdir,'config.ini')])
            # write out config file in case it doesn't exist yet
            with open(os.path.join(configdir,'config.ini'),'wb') as cf:
                config.write(cf)
        else:
            os.mkdir(configdir)
            # write out config file
            with open(os.path.join(configdir,'config.ini'),'wb') as cf:
                config.write(cf)
        # return the generated config
        return config

    # fontawesome strings and subjective interpretations of song ratings.
    def songratings(self,config):
        sr= {
            '0': {'stars': u"\uf006\uf006\uf006\uf006\uf006",'meaning':config.get('songratings','zero')},
            '1': {'stars': u"\uf123\uf006\uf006\uf006\uf006",'meaning':config.get('songratings','one')},
            '2': {'stars': u"\uf005\uf006\uf006\uf006\uf006",'meaning':config.get('songratings','two')},
            '3': {'stars': u"\uf005\uf123\uf006\uf006\uf006",'meaning':config.get('songratings','three')},
            '4': {'stars': u"\uf005\uf005\uf006\uf006\uf006",'meaning':config.get('songratings','four')},
            '5': {'stars': u"\uf005\uf005\uf123\uf006\uf006",'meaning':config.get('songratings','five')},
            '6': {'stars': u"\uf005\uf005\uf005\uf006\uf006",'meaning':config.get('songratings','six')},
            '7': {'stars': u"\uf005\uf005\uf005\uf123\uf006",'meaning':config.get('songratings','seven')},
            '8': {'stars': u"\uf005\uf005\uf005\uf005\uf006",'meaning':config.get('songratings','eight')},
            '9': {'stars': u"\uf005\uf005\uf005\uf005\uf123",'meaning':config.get('songratings','nine')},
            '10': {'stars': u"\uf005\uf005\uf005\uf005\uf005",'meaning':config.get('songratings','ten')},
            '' : {'stars': u"\uf29c", 'meaning':'No sticker set'}
        }
        return sr

    def getfontsize(self,str):
        """Method that determines font size based on text length."""
        lr = len(str)
        if lr < 33:
            rsize = '40sp'
        elif lr >= 55:
            rsize = '24sp'
        else:
            rsize = sizearray[int(round((lr-33)/21*14))]
        return rsize

    def decodeFileName(self,name):
        """Method that tries to intelligently decode a filename to handle unicode weirdness."""
        if type(name) == str:
            try:
                name = name.decode('utf8')
            except:
                name = name.decode('windows-1252')
        return name

# helper array for scaling font sizes based on text length
sizearray = ['39sp','38sp','37sp','36sp','35sp','34sp','33sp','32sp','31sp','30sp','29sp','28sp','27sp','26sp','25sp']

class ExtraSlider(Slider):
    """Class that implements some extra stuff on top of a standard slider."""

    def __init__(self,**kwargs):
        """Do normal init routine, but also register on_release event."""
        super(self.__class__,self).__init__(**kwargs)
        self.register_event_type('on_release')

    def on_release(self):
        """Override this with something you want this slider to do."""
        pass

    def on_touch_up(self, touch):
        """Check if slider is released, dispatch the on_release event if so."""
        released = super(self.__class__,self).on_touch_up(touch)
        if released:
            self.dispatch('on_release')
        return released

class ClearButton(Button):
    """A button that is clear instead of opaque."""
    pass



import kivy
kivy.require('1.10.0')
from kivy.app import App
from kivy.logger import Logger
from kmpc.mpd import MPDProtocol
from twisted.internet import protocol

class MPDFactoryProtocol(MPDProtocol):
    """Factory for MPDProtocol."""
    def connectionMade(self):
        """Call the connectionMade override."""
        if callable(self.factory.connectionMade):
            self.factory.connectionMade(self)
    def connectionLost(self, reason):
        """Call the connectionLost override."""
        if callable(self.factory.connectionLost):
            self.factory.connectionLost(self, reason)

class MPDIdleHandler(object):
    """Handler class for mpd idle command."""

    def __init__(self, protocol):
        """Set this class's protocol to the one that was passed in."""
        self.protocol = protocol

    def __call__(self, result):
        """Handle results from mpd idle command."""
        app=App.get_running_app()

        # notify various subsystems based on what changed
        for s in result:
            Logger.info('MPDIdleHandler: Changed '+format(s))
            if format(s) == 'playlist':
                # playlist was changed, ask mpd for playlist info
                self.protocol.playlistinfo().addCallback(app.root.ids.playlist_tab.populate_playlist).addErrback(app.root.ids.playlist_tab.handle_mpd_error)
                # force a reload of nextsong if playlist changes
                app.root.nextsong = None
                self.protocol.status().addCallback(app.root.update_mpd_status).addErrback(app.root.handle_mpd_error)
            elif format(s) == 'player':
                # player was changed, ask mpd for player status
                self.protocol.status().addCallback(app.root.update_mpd_status).addErrback(app.root.handle_mpd_error)
            elif format(s) == 'sticker':
                # song rating sticker was changed, ask mpd for current song rating
                self.protocol.status().addCallback(app.root.update_mpd_status).addErrback(app.root.handle_mpd_error)
                self.protocol.sticker_get('song',app.root.currfile,'rating').addCallback(app.root.update_mpd_sticker_rating).addErrback(app.root.handle_mpd_no_sticker)
            elif format(s) == 'options':
                # some playback option was changed, ask mpd for player status
                self.protocol.status().addCallback(app.root.update_mpd_status).addErrback(app.root.handle_mpd_error)
            elif format(s) == 'message':
                # an mpd message was received, ask mpd what it was
                self.protocol.readmessages().addCallback(app.root.handle_mpd_message).addErrback(app.root.handle_mpd_error)
            else:
                # default if none of the above, ask mpd for player status
                self.protocol.status().addCallback(app.root.update_mpd_status).addErrback(app.root.handle_mpd_error)

class MPDClientFactory(protocol.ClientFactory):
    """Factory for MPDClient."""
    protocol = MPDFactoryProtocol
    connectionMade = None
    connectionLost = None

    def buildProtocol(self, addr):
        """Hook up protocol and idle handler."""
        Logger.debug('MPDClientFactory: buildProtocol')
        protocol = self.protocol()
        protocol.factory = self
        protocol.idle_result = MPDIdleHandler(protocol)
        return protocol

    def clientConnectionFailed(self, connector, reason):
        """mpd connection failed for some reason."""
        Logger.error('Connection failed - goodbye!: '+format(reason))

    def clientConnectionLost(self, connector, reason):
        """mpd connection lost for some reason."""
        Logger.error('Connection lost - goodbye!: '+format(reason))



# import dependencies
from mpd import MPDProtocol
import os
import traceback
import mutagen
import io
import random
import pickle
from functools import partial
import json
import subprocess
import tempfile
import shutil
from pkg_resources import resource_filename
import musicbrainzngs

# make sure we are on an updated version of kivy
import kivy
kivy.require('1.10.0')

#install twisted reactor to interface with mpd
from kivy.support import install_twisted_reactor
install_twisted_reactor()
from twisted.internet import reactor, protocol, task, defer, threads

# import all the other kivy stuff
from kivy.config import Config
from kivy.app import App
from kivy.support import install_twisted_reactor
from kivy.logger import Logger
from kivy.graphics import Color,Rectangle
from kivy.core.image import Image as CoreImage
from kivy.metrics import Metrics, sp
from kivy.properties import BooleanProperty
from kivy.network.urlrequest import UrlRequest
from kivy.uix.widget import Widget
from kivy.uix.tabbedpanel import TabbedPanel, TabbedPanelItem
from kivy.uix.button import Button
from kivy.uix.label import Label
from kivy.uix.scrollview import ScrollView
from kivy.uix.gridlayout import GridLayout
from kivy.uix.checkbox import CheckBox
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.image import Image,AsyncImage
from kivy.uix.popup import Popup
from kivy.uix.behaviors import ButtonBehavior, FocusBehavior
from kivy.uix.recycleview.layout import LayoutSelectionBehavior
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.uix.textinput import TextInput
from kivy.clock import Clock
from kivy.lang import Builder

# import our local modules
from mpdfactory import MPDClientFactory
from extra import KmpcHelpers
from version import VERSION, VERSION_STR

# sets the location of the config folder
configdir = os.path.join(os.path.expanduser('~'),".kmpc")

# load the manager.kv file
Builder.load_file(resource_filename(__name__,os.path.join('resources','manager.kv')))

Helpers=KmpcHelpers()

class ArtistRecycleBoxLayout(FocusBehavior,LayoutSelectionBehavior,RecycleBoxLayout):
    ''' Adds selection and focus behaviour to the view. '''

class ArtistRow(RecycleDataViewBehavior,BoxLayout):
    ''' Add selection support to the Label '''
    index = None
    selected = BooleanProperty(False)
    selectable = BooleanProperty(True)

    def refresh_view_attrs(self, rv, index, data):
        ''' Catch and handle the view changes '''
        self.index = index
        return super(ArtistRow, self).refresh_view_attrs(
            rv, index, data)

    def on_touch_down(self, touch):
        ''' Add selection on touch down '''
        if super(ArtistRow, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos) and self.selectable:
            return self.parent.select_with_touch(self.index, touch)

    def apply_selection(self, rv, index, is_selected):
        ''' Respond to the selection of items in the view. '''
        self.selected = is_selected
        if is_selected:
            App.get_running_app().root.selected_row=index

class UneditTextInput(TextInput):
    def insert_text(self, substring, from_undo=False):
        pass

class ManagerInterface(TabbedPanel):

    def __init__(self,config):
        super(self.__class__,self).__init__()
        # pull config into the class
        self.config = config
        # set up mpd connection
        self.factory = MPDClientFactory()
        self.factory.connectionMade = self.mpd_connectionMade
        self.factory.connectionLost = self.mpd_connectionLost
        reactor.connectTCP(self.config.get('mpd','mpdhost'), self.config.getint('mpd','mpdport'), self.factory)
        self.artist_id_hash={}
        self.artist_name_hash={}
        self.media_hash={}
        self.wr_hash={}
        self.totaldone=0
        self.selected_row=None
        musicbrainzngs.set_useragent("kmpcmanager",VERSION_STR,'https://github.com/eratosthene/kmpc')
        self.fanarturl="http://webservice.fanart.tv/v3/music/"
        self.api_key="406b2a5af85c14b819c1c6332354b313"

    def mpd_connectionMade(self,protocol):
        self.protocol = protocol
        Logger.info('Manager: Connected to mpd server host='+self.config.get('mpd','mpdhost')+' port='+self.config.get('mpd','mpdport'))
        self.ids.library_tab.protocol = self.protocol
        self.refresh_artists_from_cache()

    def mpd_connectionLost(self,protocol, reason):
        Logger.warn('Manager: Connection lost: %s' % reason)

    def handle_mpd_error(self,result):
        Logger.error('Manager: MPDIdleHandler Callback error: {}'.format(result))

    def refresh_artists(self):
        self.protocol.list('musicbrainz_artistid').addCallback(self.populate_artists).addErrback(self.handle_mpd_error)

    def populate_artists(self,result):
        Logger.info("Manager: populate_artists")
        self.totaldone=0
        waittime=1
        self.wr_hash={}
        for row in result:
            aids = str(row)
            for aid in aids.split('/'):
                if aid not in self.artist_id_hash and aid not in self.wr_hash and len(aid)>0:
                    Logger.debug("MusicBrainz: scheduling query in "+str(waittime)+" seconds")
                    Clock.schedule_once(partial(self.query_mb,aid),waittime)
                    waittime=waittime+1
                    self.wr_hash[aid]=True

    def query_mb(self,aid,*largs):
        Logger.info("MusicBrainz: get_artist_by_id("+aid+")")
        try:
            mbres=musicbrainzngs.get_artist_by_id(aid)
        except WebServiceError as exc:
            Logger.error("MusicBrainz: web service error "+format(exc))
        else:
            aname=mbres['artist']['name']
            Logger.debug("query_mb: result from musicbrainz for aid "+aid+": "+aname)
            self.artist_id_hash[aid]=aname
            self.artist_name_hash[aname]=aid
            data = {'artist_id':aid,'artist_name':aname}
            self.ids.artist_tab.rv.data.append(data)
            self.ids.artist_tab.rv.refresh_from_data()
            self.totaldone=self.totaldone+1
            self.ids.status.text=aid+' ('+str(self.totaldone)+')'

    def write_artists_to_cache(self):
        cachefile=open(os.path.join(configdir,'artist_cache.pkl'),'w')
        pickle.dump((self.artist_id_hash,self.artist_name_hash,self.media_hash),cachefile,-1)
        cachefile.close()

    def refresh_artists_from_cache(self):
        try:
            cachefile=open(os.path.join(configdir,'artist_cache.pkl'),'r')
            (self.artist_id_hash,self.artist_name_hash,self.media_hash)=pickle.load(cachefile)
            cachefile.close()
        except IOError:
            pass
        self.ids.artist_tab.rv.data=[]
        newdata=[]
        for aid,aname in self.artist_id_hash.iteritems():
            try:
                has_artistbackground=self.media_hash[aid]['has_artistbackground']
            except:
                has_artistbackground=False
            try:
                has_logo=self.media_hash[aid]['has_logo']
            except:
                has_logo=False
            try:
                has_badge=self.media_hash[aid]['has_badge']
            except:
                has_badge=False
            datum = {'artist_id':aid,'artist_name':aname,'has_artistbackground':has_artistbackground,'has_logo':has_logo,'has_badge':has_badge}
            newdata.append(datum)
        self.ids.status.text='pulled '+str(len(self.artist_id_hash))+' lines from cache'
        self.ids.artist_tab.rv.data=sorted(newdata,key=lambda k: k['artist_name'])

    def scan_for_media(self,index):
        Logger.info('Manager: scanning '+self.ids.artist_tab.rv.data[index]['artist_id']+'for media')
        fa_path=self.config.get('paths','fanartpath')
        artistbackground_path=os.path.join(fa_path,self.ids.artist_tab.rv.data[index]['artist_id'],'artistbackground')
        logo_path=os.path.join(fa_path,self.ids.artist_tab.rv.data[index]['artist_id'],'logo')
        badge_path=os.path.join(fa_path,self.ids.artist_tab.rv.data[index]['artist_id'],'badge')
        self.ids.artist_tab.rv.data[index]['has_artistbackground']=os.path.isdir(artistbackground_path)
        self.ids.artist_tab.rv.data[index]['has_logo']=os.path.isdir(logo_path)
        self.ids.artist_tab.rv.data[index]['has_badge']=os.path.isdir(badge_path)
        self.media_hash[self.ids.artist_tab.rv.data[index]['artist_id']]={}
        self.media_hash[self.ids.artist_tab.rv.data[index]['artist_id']]['has_artistbackground']=os.path.isdir(artistbackground_path)
        self.media_hash[self.ids.artist_tab.rv.data[index]['artist_id']]['has_logo']=os.path.isdir(logo_path)
        self.media_hash[self.ids.artist_tab.rv.data[index]['artist_id']]['has_badge']=os.path.isdir(badge_path)
        self.ids.artist_tab.rv.refresh_from_data()

    def scan_row_for_media(self):
        if self.selected_row is not None:
            self.scan_for_media(self.selected_row)
        self.write_artists_to_cache()

    def scan_all_for_media(self):
        for idx in range(0,len(self.ids.artist_tab.rv.data)):
            self.scan_for_media(idx)
        self.write_artists_to_cache()

    def trim_image(self,filename,request,result):
        tdir=tempfile.mkdtemp()
        tf1=os.path.join(tdir,'tf1.png')
        Logger.debug("trim_image: fixing "+filename)
        Logger.debug("trim_image: using tfile "+tf1)
        subprocess.call(['convert',filename,'-bordercolor','none','-border','10x10',tf1])
        subprocess.call(['convert',tf1,'-trim','+repage',filename])
        shutil.rmtree(tdir)

    def pull_art(self,index):
        Logger.info('Manager: pulling art for '+self.ids.artist_tab.rv.data[index]['artist_id'])
        aid=self.ids.artist_tab.rv.data[index]['artist_id']
        aname=self.ids.artist_tab.rv.data[index]['artist_name']
        fa_path=self.config.get('paths','fanartpath')
        fanart=self.fanarturl
        api_key=self.api_key
        furl=fanart+aid+"?api_key="+api_key
        client_key=self.config.get('fanart','client_key')
        if client_key:
            furl=furl+"&client_key="+client_key
        Logger.debug("pull_art: querying "+furl)
        request = UrlRequest(url=furl,on_success=partial(self.pull_art2,index))

    def pull_art2(self,index,request,result):
        aid=self.ids.artist_tab.rv.data[index]['artist_id']
        aname=self.ids.artist_tab.rv.data[index]['artist_name']
        fa_path=self.config.get('paths','fanartpath')
        d=result
        if 'hdmusiclogo' in d or 'artistbackground' in d or 'musiclogo' in d:
            fapath=os.path.join(fa_path,aid)
            lpath=os.path.join(fapath,"logo")
            abpath=os.path.join(fapath,"artistbackground")
            bpath=os.path.join(fapath,"badge")
            try:
                Logger.debug("pull_art2: downloading to "+fapath)
                os.mkdir(fapath)
                with open(os.path.join(fapath,"__"+aname.replace(os.sep,'_')+"__"),'w'):
                    pass
            except OSError:
                pass
            if 'hdmusiclogo' in d:
                try:
                    os.mkdir(lpath)
                except OSError:
                    pass
                for idx,img in enumerate(d['hdmusiclogo']):
                    if not os.path.isfile(os.path.join(lpath,img['id']+'.png')) and not os.path.isfile(os.path.join(bpath,img['id']+'.png')):
                        Logger.debug("pull_art2: downloading hdmusiclogo "+img['id'])
                        fp=os.path.join(lpath,img['id']+'.png')
                        req = UrlRequest(img['url'],on_success=partial(self.trim_image,fp),file_path=fp)
                        if self.config.getboolean('logs','artlog'):
                            adfile=open(os.path.join(configdir,'artlog.txt'),'a')
                            adfile.write(os.path.join(lpath,img['id']+'.png')+"\n")
                            adfile.close()
            if 'musiclogo' in d:
                try:
                    os.mkdir(lpath)
                except OSError:
                    pass
                for idx,img in enumerate(d['musiclogo']):
                    if not os.path.isfile(os.path.join(lpath,img['id']+'.png')) and not os.path.isfile(os.path.join(bpath,img['id']+'.png')):
                        Logger.debug("pull_art2: downloading musiclogo "+img['id'])
                        fp=os.path.join(lpath,img['id']+'.png')
                        req = UrlRequest(img['url'],on_success=partial(self.trim_image,fp),file_path=fp)
                        if self.config.getboolean('logs','artlog'):
                            adfile=open(os.path.join(configdir,'artlog.txt'),'a')
                            adfile.write(os.path.join(lpath,img['id']+'.png')+"\n")
                            adfile.close()
            if 'artistbackground' in d:
                try:
                    os.mkdir(abpath)
                except OSError:
                    pass
                for idx,img in enumerate(d['artistbackground']):
                    if not os.path.isfile(os.path.join(abpath,img['id']+'.png')):
                        Logger.debug("pull_art2: downloading artistbackground "+img['id'])
                        fp=os.path.join(abpath,img['id']+'.png')
                        req = UrlRequest(img['url'],file_path=fp)
                        if self.config.getboolean('logs','artlog'):
                            adfile=open(os.path.join(configdir,'artlog.txt'),'a')
                            adfile.write(os.path.join(abpath,img['id']+'.png')+"\n")
                            adfile.close()

    def pull_art_for_row(self):
        if self.selected_row is not None:
            self.pull_art(self.selected_row)
            self.scan_for_media(self.selected_row)
        self.write_artists_to_cache()

    def pull_art_for_all(self):
        for idx in range(0,len(self.ids.artist_tab.rv.data)):
            datum=self.ids.artist_tab.rv.data[idx]
            try:
                if not datum['has_logo'] and not datum['has_badge'] and not datum['has_artistbackground']:
                    self.pull_art(idx)
                    self.scan_for_media(idx)
            except KeyError:
                pass
        self.write_artists_to_cache()


class ManagerApp(App):

    def __init__(self,args):
        Config.set('graphics','width',1280)
        Config.set('graphics','height',720)
        Config.set('kivy','keyboard_mode','system')
        super(self.__class__,self).__init__()

    def build(self):
        config=Helpers.loadconfigfile()
        # setup some variables that interface.kv will use
        # this is necessary to support packaging the app
        self.songratings = Helpers.songratings(config)
        self.normalfont = resource_filename(__name__,os.path.join('resources','DejaVuSans.ttf'))
        self.fontawesomefont = resource_filename(__name__,os.path.join('resources','FontAwesome.ttf'))
        return ManagerInterface(config)

if __name__ == '__main__':
    ManagerApp().run()




# import dependencies
from mpd import MPDProtocol
import os
import traceback
import mutagen
import io
import random
import ConfigParser
from pkg_resources import resource_filename

# make sure we are on an updated version of kivy
import kivy
kivy.require('1.10.0')

#install twisted reactor to interface with mpd
from kivy.support import install_twisted_reactor
install_twisted_reactor()
from twisted.internet import reactor, protocol
from twisted.internet.defer import inlineCallbacks

# import all the other kivy stuff
from kivy.config import Config
from kivy.app import App
from kivy.logger import Logger
from kivy.graphics import Color,Rectangle
from kivy.core.image import Image as CoreImage
from kivy.metrics import Metrics, sp
from kivy.clock import Clock
from kivy.uix.widget import Widget
from kivy.uix.tabbedpanel import TabbedPanel, TabbedPanelItem
from kivy.uix.button import Button
from kivy.uix.label import Label
from kivy.uix.scrollview import ScrollView
from kivy.uix.gridlayout import GridLayout
from kivy.uix.checkbox import CheckBox
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.image import Image,AsyncImage
from kivy.uix.popup import Popup
from kivy.uix.behaviors import ButtonBehavior
from kivy.uix.slider import Slider
from kivy.factory import Factory
from kivy.lang import Builder

# import our local modules
from mpdfactory import MPDClientFactory
from extra import KmpcHelpers,ExtraSlider,ClearButton
from playlistpanel import PlaylistTabbedPanelItem

# sets the location of the config folder
configdir = os.path.join(os.path.expanduser('~'),".kmpc")

# load the interface.kv file
Builder.load_file(resource_filename(__name__,os.path.join('resources','interface.kv')))

Helpers=KmpcHelpers()

class KmpcInterface(TabbedPanel):
    """The main class that ties it all together."""

    def __init__(self):
        """Pull the config from the config file, hook up to mpd, zero out variables."""
        super(self.__class__,self).__init__()
        # pull config into the class
        self.config = Helpers.loadconfigfile()
        # set up mpd connection
        self.factory = MPDClientFactory()
        self.factory.connectionMade = self.mpd_connectionMade
        self.factory.connectionLost = self.mpd_connectionLost
        reactor.connectTCP(self.config.get('mpd','mpdhost'), self.config.getint('mpd','mpdport'), self.factory)
        # bind callbacks for tab changes
        self.bind(current_tab=self.main_tab_changed)
        self.mpd_status={'state':'stop','repeat':0,'single':0,'random':0,'consume':0,'curpos':0}
        self.currsong=None
        self.nextsong=None
        self.currfile=None
        self.track_slider_task=None
        self.accessoryPopup=Factory.AccessoryPopup()

    def mpd_connectionMade(self,protocol):
        """Callback when mpd is connected."""
        # copy the protocol to all the classes
        self.protocol = protocol
        self.ids.playlist_tab.protocol=protocol
        self.ids.library_tab.protocol=protocol
        self.ids.config_tab.protocol=protocol
        Logger.info('Application: Connected to mpd server host='+self.config.get('mpd','mpdhost')+' port='+self.config.get('mpd','mpdport'))
        # get the initial mpd status
        self.protocol.status().addCallback(self.update_mpd_status).addErrback(self.handle_mpd_error)
        # create the once-per-second update of the track slider
        self.track_slider_task=Clock.schedule_interval(self.update_track_slider,1)
        self.track_slider_task.cancel()
        # subscribe to 'kmpc' to check for messages from mpd
        self.protocol.subscribe('kmpc')

    def main_tab_changed(self,obj,value):
        """Callback when top tab is changed."""
        self.active_tab = value.text
        Logger.info("Application: Changed active tab to "+self.active_tab)
        if self.active_tab == 'Now Playing':
            pass
        elif self.active_tab == 'Playlist':
            pass
            # switching to the playlist tab repopulates it if it is empty
            # this is skipped right now by the 'pass' above, I can't remember why, I think it's handled a different way now
            if len(self.ids.playlist_tab.rv.data) == 0:
                self.protocol.playlistinfo().addCallback(self.ids.playlist_tab.populate_playlist).addErrback(self.ids.playlist_tab.handle_mpd_error)
        elif self.active_tab == 'Config':
            # switching to the config tab repopulates options
            self.protocol.status().addCallback(self.ids.config_tab.update_mpd_status).addErrback(self.ids.config_tab.handle_mpd_error)

    def mpd_connectionLost(self,protocol, reason):
        """Callback when mpd connection is lost."""
        Logger.warn('Application: Connection lost: %s' % reason)
        # kills the app for now since I don't know how to handle this yet
        App.get_running_app().stop()

    def handle_mpd_error(self,result):
        """Prints handled errors to the error log."""
        Logger.error('Application: MPDIdleHandler Callback error: {}'.format(result))

    def handle_mpd_message(self,result):
        """Callback for mpd 'kmpc' messages."""
        # result is an iterator although it only generally has one member
        for m in result:
            message = m['message']
            Logger.info("Application: MPD Message: "+message)
            if message == 'AccessoryOff':
                # generated by mausberry circuit, accessory power is off
                self.accessoryPopup.open()
            elif message == 'AccessoryOn':
                # generated by mausberry circuit, accessory power is on
                self.accessoryPopup.dismiss()

    def current_track_slider_release(self):
        """Callback when track slider is released."""
        # get the value of the track slider
        curpos = int(self.ids.current_track_slider.value)
        Logger.debug("Application: touch up on track slider at "+str(curpos))
        # seek to that position in the track
        self.protocol.seekcur(str(curpos)).addErrback(self.handle_mpd_error)

    def current_track_slider_down(self):
        """Callback when track slider is pressed."""
        Logger.debug("Application: touch down on track slider")
        # cancel the once-per-second update while finger is down so it doesn't move out from under it
        self.track_slider_task.cancel()

    def update_track_slider(self,dt):
        """Increment the slider by 1 second every second."""
        Logger.debug("Application: update_track_slider")
        curpos=int(self.ids.current_track_slider.value)+1
        self.ids.current_track_slider.value=curpos

    def stop_zero_stuff(self):
        """Clear the screen and variables if playback is fully stopped."""
        self.ids.current_track_slider.value=0
        self.ids.current_track_slider.max=0
        self.ids.current_playlist_track_number_label.text=''
        self.ids.next_song_artist_label.text = ''
        self.currfile = None
        self.currsong = None
        self.nextsong = None
        self.ids.song_star_layout.clear_widgets()
        self.ids.album_cover_layout.clear_widgets()
        self.ids.trackinfo.clear_widgets()
        lbl = Label(text="Playback Stopped")
        self.ids.trackinfo.add_widget(lbl)
        self.ids.player.canvas.before.add(Rectangle(source=resource_filename(__name__,os.path.join("resources","backdrop.png")),size=self.ids.player.size,pos=self.ids.player.pos))

    def update_mpd_status(self,result):
        """Callback when mpd status changes."""
        Logger.debug('NowPlaying: update_mpd_status()')
        # probably state is the only one necessary, but hey
        self.mpd_status['state']=result['state']
        self.mpd_status['repeat']=result['repeat']
        self.mpd_status['single']=result['single']
        self.mpd_status['random']=result['random']
        self.mpd_status['consume']=result['consume']
        if self.mpd_status['state'] == 'stop':
            # if stopped, there are no current or next songs
            self.currsong=None
            self.nextsong=None
        else:
            # save current song, this is a 0-based index into the playlist
            self.currsong=result['song']
            # ask mpd for current song data
            self.protocol.currentsong().addCallback(self.update_mpd_currentsong).addErrback(self.handle_mpd_error)
            # save next song, this is a 0-based index into the playlist
            if 'nextsong' in result:
                self.nextsong=result['nextsong']
                # ask mpd for next song data
                self.protocol.playlistinfo(self.nextsong).addCallback(self.update_mpd_nextsong).addErrback(self.handle_mpd_error)
            else:
                self.nextsong=None
        stflags=['normal','down']
        # check various flag states
        self.ids.repeat_button.state=stflags[int(self.mpd_status['repeat'])]
        self.ids.single_button.state=stflags[int(self.mpd_status['single'])]
        self.ids.random_button.state=stflags[int(self.mpd_status['random'])]
        self.ids.consume_button.state=stflags[int(self.mpd_status['consume'])]
        if self.mpd_status['state']=='pause' or self.mpd_status['state']=='stop':
            # play/pause button should be a play button that is unpressed
            self.ids.play_button.state='normal'
            self.ids.play_button.text=u"\uf04b"
            self.track_slider_task.cancel()
        else:
            # play/pause button should be a pause button that is pressed
            self.ids.play_button.state='down'
            self.ids.play_button.text=u"\uf04c"
        if self.mpd_status['state'] == 'stop':
            # zero everything out if we are stopped
            self.stop_zero_stuff()
        else:
            # mpd returns {elapsed seconds}:{total seconds}, the following splits each to minute:second
            c,t=result['time'].split(":")
            # set the max slider value to the total seconds
            self.ids.current_track_slider.max = int(t)
            # set the current slider value to the current seconds
            self.mpd_status['curpos']=int(c)
            self.ids.current_track_slider.value = int(c)
            if self.mpd_status['state']!='pause':
                self.track_slider_task()
            # throws an exception if i don't do this
            a=int(result['song'])+1
            b=int(result['playlistlength'])
            self.ids.current_playlist_track_number_label.text = "%d of %d" % (a,b)
        # update the playlist tab with status results
        self.ids.playlist_tab.update_mpd_status(result)
        # update the config tab with status results
        self.ids.config_tab.update_mpd_status(result)

    def update_mpd_currentsong(self,result):
        """Callback for mpd currentsong data."""
        Logger.debug('NowPlaying: update_mpd_currentsong()')
        # this is so expensive screen updates only happen if the song has changed since the last time this callback was called
        songchange=False
        # ti is the track info widget
        ti=self.ids.trackinfo
        # if result is undefined, there's not actually a song playing
        if result:
            # if class's current file doesn't match what mpd returns, the song has changed
            if self.currfile != result['file']:
                songchange=True
            # update class's current file
            self.currfile = result['file']
            if songchange:
                # clear the track info widget
                ti.clear_widgets()
                # clear the album cover
                self.ids.album_cover_layout.clear_widgets()
                # get the stored star rating
                self.protocol.sticker_get('song',self.currfile,'rating').addCallback(self.update_mpd_sticker_rating).addErrback(self.handle_mpd_no_sticker)
                # figure out the full path of the file
                bp=self.config.get('paths','musicpath')
                # p is the absolute path
                p=os.path.join(bp,result['file'])
                haslogo=False
                # set the release year if mpd has it
                year=None
                if 'date' in result:
                    year=result['date'][:4]
                # pull the fanart folder from ini file
                fa_path=self.config.get('paths','fanartpath')
                # try to get the artistid, set it to 0000 if it doesn't exist
                try:
                    aids=str(result['musicbrainz_artistid'])
                except KeyError:
                    aids='0000'
                # if a track has multiple artists, split it up
                artistids=aids.split('/')
                # create a widget to hold the cover art
                cbl = BoxLayout(size_hint=(1,1),orientation='horizontal')
                # loop through artistids
                for mb_aid in artistids:
                    # try to find a logo for this artist in the fanart folder
                    try:
                        # look in the 'logo' subfolder of the artistid path
                        al_path=os.path.join(fa_path,mb_aid,'logo')
                        # pick one at random
                        img_path=os.path.join(al_path,random.choice(os.listdir(al_path)))
                        if os.path.isfile(img_path):
                            # create an image button out of the logo so you can press it
                            img = ImageButton(source=os.path.join(al_path,img_path),allow_stretch=True,color=(1,1,1,0.65))
                            cbl.add_widget(img)
                            haslogo=True
                    except:
                        pass
                if haslogo:
                    # we found a logo, show it
                    current_artist_label = cbl
                else:
                    # no logo found, just print the artist name instead
                    current_artist_label = InfoLargeLabel(text = result['artist'],font_size=Helpers.getfontsize(result['artist']))
                # clear the background of the player canvas so we can show the artist background
                self.ids.player.canvas.before.clear()
                # pick an artistid at random for the background
                mb_aid=random.choice(artistids)
                try:
                    # look in the 'artistbackground' subfolder of the artistid path
                    ab_path=os.path.join(fa_path,mb_aid,'artistbackground')
                    # pick one at random
                    img_path=random.choice(os.listdir(ab_path))
                    # update the player background with the image
                    self.ids.player.canvas.before.add(Rectangle(source=os.path.join(ab_path,img_path),size=self.ids.player.size,pos=self.ids.player.pos))
                except:
                    # update the player background with the default backdrop
                    self.ids.player.canvas.before.add(Rectangle(source=resource_filename(__name__,os.path.join("resources","backdrop.png")),size=self.ids.player.size,pos=self.ids.player.pos))
                if os.path.isfile(p):
                    Logger.debug('NowPlaying: found good file at path '+p)
                    # load up the file to read the tags
                    f = mutagen.File(p)
                    cimg = None
                    data = None
                    # if the original year mp3 tag exists use it instead of mpd's year
                    # I prefer this year to be displayed, rather than the year an album was remastered
                    if 'TXXX:originalyear' in f.keys():
                        year=format(f['TXXX:originalyear'])
                    # try to get mp3 cover, if this throws an exception it's not an mp3 or it doesn't have a cover
                    try:
                        pframes = f.tags.getall("APIC")
                        # id3v2 can store any number of image frames, we just want the first one
                        for frame in pframes:
                            ext = 'img'
                            # figure out the file type
                            if frame.mime.endswith('jpeg') or frame.mime.endswith('jpg'):
                                ext = 'jpg'
                            elif frame.mime.endswith('png'):
                                ext = 'png'
                            elif frame.mime.endswith('bmp'):
                                ext = 'bmp'
                            elif frame.mime.endswith('gif'):
                                ext = 'gif'
                            # pull the raw image data into a variable
                            data=io.BytesIO(bytearray(frame.data))
                            break
                    except AttributeError:
                        pass
                    # try to get mp4 cover
                    if 'covr' in f.keys():
                        # figure out the file type
                        if f['covr'][0].imageformat == mutagen.mp4.MP4Cover.FORMAT_PNG:
                            ext = 'png'
                        else:
                            ext = 'jpg'
                        # pull the raw image data into a variable
                        data=io.BytesIO(bytearray(f['covr'][0]))
                    if data:
                        # if we got image data, load it as a kivy.core.image
                        cimg = CoreImage(data,ext=ext)
                    if cimg:
                        # if the image loading worked, create an image widget and fix up the layout
                        # TODO: sometimes this just returns a black rectangle, i think i need to catch more specific exceptions
                        # and figure out what exactly is happening
                        img=ImageButton(texture=cimg.texture,allow_stretch=True)
                        self.ids.album_cover_layout.add_widget(img)
                        # popup the cover large if you press it
                        img.bind(on_press=self.cover_popup)
                else:
                    # this should _probably_ never happen
                    Logger.debug('NowPlaying: no file found at path '+p)
                # add the correct artist name widget
                ti.add_widget(current_artist_label)
                # if we got a year tag from somewhere, include it in the album label
                if year:
                    yeartext = result['album']+' ['+year+']'
                else:
                    yeartext = result['album']
                if haslogo:
                    # we found an artist logo, put the song and album labels in a separate boxlayout to separate them a bit
                    lyt = BoxLayout(orientation='vertical',padding_y='10sp')
                    current_song_label = InfoLargeLabel(text = result['title'],font_size=Helpers.getfontsize(result['title']))
                    lyt.add_widget(current_song_label)
                    current_album_label = InfoLargeLabel(text = yeartext,font_size=Helpers.getfontsize(yeartext))
                    lyt.add_widget(current_album_label)
                    ti.add_widget(lyt)
                else:
                    # no artist logo, just add the song and album labels directly to the track info widget
                    current_song_label = InfoLargeLabel(text = result['title'],font_size=Helpers.getfontsize(result['title']))
                    ti.add_widget(current_song_label)
                    current_album_label = InfoLargeLabel(text = yeartext,font_size=Helpers.getfontsize(yeartext))
                    ti.add_widget(current_album_label)
        else:
            # there's not a current song, so zero everything out
            self.stop_zero_stuff()

    def update_mpd_sticker_rating(self,result):
        """Callback for song that has a rating in mpd."""
        Logger.debug('NowPlaying: update_mpd_sticker_rating')
        # make a clear button for the star rating
        btn = ClearButton(padding_x='10sp',font_name=resource_filename(__name__,os.path.join('resources','FontAwesome.ttf')),halign='center',valign='middle',markup=True)
        # look up the correct string for the rating
        btn.text = Helpers.songratings(self.config)[result]['stars']
        # bind the popup for setting rating
        btn.bind(on_press=self.rating_popup)
        # clear the layout widget and add the new one
        self.ids.song_star_layout.clear_widgets()
        self.ids.song_star_layout.add_widget(btn)

    def handle_mpd_no_sticker(self,result):
        """Callback for song that has no rating in mpd."""
        Logger.debug('NowPlaying: handle_mpd_no_sticker')
        # make a clear button for the star rating
        btn = ClearButton(padding_x='10sp',font_name=resource_filename(__name__,os.path.join('resources','FontAwesome.ttf')),halign='center',valign='middle',markup=True)
        # set the string to the circled question mark icon
        btn.text = u"\uf29c"
        # bind the popup for setting rating
        btn.bind(on_press=self.rating_popup)
        # clear the layout widget and add the new one
        self.ids.song_star_layout.clear_widgets()
        self.ids.song_star_layout.add_widget(btn)

    def update_mpd_nextsong(self,result):
        """Callback for next song data from mpd."""
        Logger.debug('NowPlaying: update_mpd_nextsong()')
        # result is a list with one member
        for obj in result:
            # set the next song label
            self.ids.next_song_artist_label.text = 'Up Next: '+obj['artist']+' - '+obj['title']

    def prev_pressed(self):
        """Callback for prev button pressed."""
        Logger.debug('Application: prev_pressed()')
        self.protocol.previous()

    def play_pressed(self):
        """Callback for play/pause button pressed."""
        Logger.debug('Application: play_pressed()')
        if self.mpd_status['state'] == 'play':
            # pause if playing
            self.protocol.pause()
        else:
            # play if paused or stopped
            self.protocol.play()

    def next_pressed(self):
        """Callback for next button pressed."""
        Logger.debug('Application: next_pressed()')
        self.protocol.next()

    def repeat_pressed(self):
        """Callback for repeat button pressed."""
        Logger.debug('Application: repeat_pressed()')
        # toggle on/off
        self.protocol.repeat(str(1-int(self.mpd_status['repeat'])))

    def single_pressed(self):
        """Callback for single button pressed."""
        Logger.debug('Application: single_pressed()')
        # toggle on/off
        self.protocol.single(str(1-int(self.mpd_status['single'])))

    def random_pressed(self):
        """Callback for random button pressed."""
        Logger.debug('Application: random_pressed()')
        # toggle on/off
        self.protocol.random(str(1-int(self.mpd_status['random'])))

    def consume_pressed(self):
        """Callback for consume button pressed."""
        Logger.debug('Application: consume_pressed()')
        # toggle on/off
        self.protocol.consume(str(1-int(self.mpd_status['consume'])))

    def rating_popup(self,instance):
        """Popup for setting song rating."""
        Logger.debug('Application: rating_popup()')
        # create a layout and add it to the popup
        layout = GridLayout(cols=2,spacing=10)
        popup = Popup(title='Rating',content=layout,size_hint=(0.8,1))
        # loop from 0-10
        for r in list(range(0,11)):
            # make a button
            btn=Button(font_name=resource_filename(__name__,os.path.join('resources','FontAwesome.ttf')))
            # look up the correct string for the rating
            btn.text=Helpers.songratings(self.config)[str(r)]['stars']
            # set some widget variables
            btn.rating=str(r)
            btn.popup=popup
            # add the button to the layout
            layout.add_widget(btn)
            # bind the button press to set the rating
            btn.bind(on_press=self.rating_set)
            # add a label to explain the ratings, this is pretty subjective
            lbl=Label(text=Helpers.songratings(self.config)[str(r)]['meaning'],halign='left')
            # add the label to the layout
            layout.add_widget(lbl)
        # pop it on up, if you press outside the popup it just goes away without setting the rating
        popup.open()

    def rating_set(self,instance):
        """Method that sets a song's rating."""
        Logger.debug('Application: rating_set('+instance.rating+')')
        # close the rating popup
        instance.popup.dismiss()
        # tell mpd to set the rating sticker
        self.protocol.sticker_set('song',self.currfile,'rating',instance.rating)

    def cover_popup(self,instance):
        """Popup for showing a larger version of the album cover."""
        Logger.debug('Application: cover_popup()')
        layout = BoxLayout()
        popup = Popup(title='Cover',content=layout,size_hint=(0.6,1))
        # pull the already loaded image texture
        img = Image(texture=instance.texture,allow_stretch=True)
        # add it to the layout
        layout.add_widget(img)
        # pop pop pop
        popup.open()

    def change_backlight(self,value):
        """Method that sets the backlight to a certain value."""
        Logger.info('Application: change_backlight('+str(value)+') rpienable = '+self.config.get('flags','rpienable'))
        # only if the ini file says it's ok
        if self.config.getboolean('flags','rpienable'):
            import rpi_backlight as bl
            # set the brightness
            bl.set_brightness(int(value), smooth=True, duration=1)

class KmpcApp(App):
    """The overall app class, builds the main interface widget."""

    def __init__(self,args):
        """Override kivy config values with necessary ones."""
        Config.set('kivy','keyboard_mode','systemanddock')
        Config.set('graphics','width',800)
        Config.set('graphics','height',480)
        super(self.__class__,self).__init__()

    def build(self):
        """Instantiates KmpcInterface."""
        # setup some variables that interface.kv will use
        # this is necessary to support packaging the app
        self.normalfont = resource_filename(__name__,os.path.join('resources','DejaVuSans.ttf'))
        self.boldfont = resource_filename(__name__,os.path.join('resources','DejaVuSans-Bold.ttf'))
        self.fontawesomefont = resource_filename(__name__,os.path.join('resources','FontAwesome.ttf'))
        self.buttonnormal = resource_filename(__name__,os.path.join('resources','button-normal.png'))
        self.buttondown = resource_filename(__name__,os.path.join('resources','button-down.png'))
        self.clear = resource_filename(__name__,os.path.join('resources','clear.png'))
        self.backdrop = resource_filename(__name__,os.path.join('resources','backdrop.png'))
        self.listbackdrop = resource_filename(__name__,os.path.join('resources','list-backdrop.png'))
        self.listbackdropselected = resource_filename(__name__,os.path.join('resources','list-backdrop-selected.png'))
        self.trackslidercursor = resource_filename(__name__,os.path.join('resources','track-slider-cursor.png'))
        return KmpcInterface()

class InfoLargeLabel(Label):
    """A label with large text."""
    pass

class InfoSmallLabel(Label):
    """A label with small text."""
    pass

class ImageButton(ButtonBehavior, AsyncImage):
    """An image that you can press."""
    pass

if __name__ == '__main__':
    # run the app!
    KmpcApp().run()




import kivy
kivy.require('1.10.0')
from kivy.app import App
from kivy.uix.tabbedpanel import TabbedPanelItem
from kivy.uix.button import Button
from kivy.uix.label import Label
from kivy.uix.gridlayout import GridLayout
from kivy.uix.checkbox import CheckBox
from kivy.logger import Logger
from twisted.internet.defer import inlineCallbacks
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.properties import BooleanProperty
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.uix.behaviors import FocusBehavior
from kivy.uix.recycleview.layout import LayoutSelectionBehavior
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.popup import Popup
from copy import deepcopy
from functools import partial
import os
from pkg_resources import resource_filename

from extra import KmpcHelpers

Helpers=KmpcHelpers()

# sets the location of the config folder
configdir = os.path.join(os.path.expanduser('~'),".kmpc")

class ManagerLibraryTabbedPanelItem(TabbedPanelItem):
    current_view = {'value': 'root', 'base':'/','info':{'type':'uri'}}
    library_selection = {}
    rsync_data={}
    rsync_file=None

    def change_view_type(self,value):
        Logger.info("Library: View changed to "+value)
        self.rbl.clear_selection()
        if value == 'Files':
            self.current_view = {'value': 'root','base':'/','info':{'type':'uri'}}
            self.protocol.lsinfo(self.current_view['base']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
            self.ids.files_button.state='down'
            self.ids.albums_button.state='normal'
            self.ids.tracks_button.state='normal'
            self.ids.playlists_button.state='normal'
        elif value == 'Albums':
            self.current_view = {'value': 'All Album Artists','base':'All Album Artists','info':{'type':'rootalbums'}}
            self.protocol.list('albumartistsort').addCallback(self.reload_view).addErrback(self.handle_mpd_error)
            self.ids.files_button.state='normal'
            self.ids.albums_button.state='down'
            self.ids.tracks_button.state='normal'
            self.ids.playlists_button.state='normal'
        elif value == 'Tracks':
            self.current_view = {'value': 'All Track Artists','base':'All Track Artists','info':{'type':'roottracks'}}
            self.protocol.list('artistsort').addCallback(self.reload_view).addErrback(self.handle_mpd_error)
            self.ids.files_button.state='normal'
            self.ids.albums_button.state='normal'
            self.ids.tracks_button.state='down'
            self.ids.playlists_button.state='normal'
        elif value == 'Playlists':
            self.current_view = {'value':'All Playlists','base':'All Playlists','info':{'type':'playlist'}}
            self.protocol.listplaylists().addCallback(self.reload_view).addErrback(self.handle_mpd_error)
            self.ids.files_button.state='normal'
            self.ids.albums_button.state='normal'
            self.ids.tracks_button.state='normal'
            self.ids.playlists_button.state='down'

    def render_row(self,r,has_sticker,result):
        rr=deepcopy(r)
        if has_sticker:
            rr['copy_flag']=str(result)
        else:
            rr['copy_flag']=''
        self.protocol.sticker_get('song',rr['base'],'rating').addCallback(partial(self.render_row2,rr,True)).addErrback(partial(self.render_row2,rr,False))

    def render_row2(self,r,has_sticker,result):
        rr=deepcopy(r)
        if has_sticker:
            rr['rating']=str(result)
        else:
            rr['rating']=''
        self.rv.data.append(rr)

    def reload_view(self,result):
        Logger.info("Library: reload_view() current type: "+self.current_view['info']['type'])
        self.rv.data=[]
        if self.current_view['info']['type'] == 'uri':
            if self.current_view['base'] != '/':
                (hbase,tbase)=os.path.split(self.current_view['base'])
                (b1,b2)=os.path.split(hbase)
                if b2 == '':
                    b2 = 'root'
                upbase=os.path.normpath(self.current_view['base']+os.sep+'..')
                if upbase == '.':
                    upbase = '/'
                r = {'value':"up to "+b2,'base':upbase,'info':{'type':'uri'},'copy_flag':'','rating':''}
                self.rv.data.append(r)
                self.current_header.text = tbase
            else:
                self.current_header.text = 'All Files'
        elif self.current_view['info']['type'] == 'albumartistsort':
            r={'value': 'up to All Album Artists','base':'All Album Artists','info':{'type':'rootalbums'},'copy_flag':'','rating':''}
            self.rv.data.append(r)
            self.current_header.text = self.current_view['base']
        elif self.current_view['info']['type'] == 'album':
            r={'value':'up to '+self.current_view['info']['albumartistsort'],'base':self.current_view['info']['albumartistsort'],'info':{'type':'albumartistsort'},'copy_flag':'','rating':''}
            self.rv.data.append(r)
            self.current_header.text = self.current_view['base']
        elif self.current_view['info']['type'] == 'artistsort':
            r={'value': 'up to All Track Artists','base':'All Track Artists','info':{'type':'roottracks'},'copy_flag':'','rating':''}
            self.rv.data.append(r)
            self.current_header.text = self.current_view['base']
        else:
            self.current_header.text = self.current_view['base']
        for row in result:
            if 'playlist' in row:
                if self.current_view['info']['type'] != 'uri':
                    Logger.debug("Library: playlist found = "+row['playlist'])
                    r = {'value':row['playlist'],'base':row['playlist'],'info':{'type':'playlist'},'copy_flag':'','rating':''}
                    self.rv.data.append(r)
            elif 'directory' in row:
                Logger.debug("Library: directory found: ["+row['directory']+"]")
                (b1,b2)=os.path.split(row['directory'])
                r={'value':b2,'base':row['directory'],'info':{'type':'uri'},'copy_flag':'','rating':''}
                self.rv.data.append(r)
            elif 'file' in row:
                Logger.debug("FileBrowser: file found: ["+row['file']+"]")
                r={'value':Helpers.formatsong(row),'base':row['file'],'info':{'type':'file'}}
                self.protocol.sticker_get('song',row['file'],'copy_flag').addCallback(partial(self.render_row,r,True)).addErrback(partial(self.render_row,r,False))
            else:
                if self.current_view['info']['type'] == 'rootalbums':
                    Logger.debug("Library: album artist found: ["+row+"]")
                    r={'value':row,'base':row,'info':{'type':'albumartistsort'},'copy_flag':'','rating':''}
                    self.rv.data.append(r)
                elif self.current_view['info']['type'] == 'albumartistsort':
                    Logger.debug("Library: album found: ["+row+"]")
                    r={'value':row,'base':row,'info':{'type':'album','albumartistsort':self.current_view['base']},'copy_flag':'','rating':''}
                    self.rv.data.append(r)
                elif self.current_view['info']['type'] == 'roottracks':
                    Logger.debug("Library: track artist found: ["+row+"]")
                    r={'value':row,'base':row,'info':{'type':'artistsort'},'copy_flag':'','rating':''}
                    self.rv.data.append(r)
                elif self.current_view['info']['type'] == 'artistsort':
                    Logger.debug("Library: track found: ["+row+"]")
                    r={'value':row,'base':row,'info':{'type':'track','artistsort':self.current_view['base']},'copy_flag':'','rating':''}
                    self.rv.data.append(r)
                else:
                    Logger.warn("Library: not sure what to do with ["+format(row)+"]")

    def handle_double_click(self,row,index):
        Logger.debug("Library: handle_double_click("+format(row)+")")
        self.current_view = deepcopy(row)
        if row['info']['type'] == 'uri':
            self.protocol.lsinfo(row['base']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'rootalbums':
            self.protocol.list('albumartistsort').addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'albumartistsort':
            self.protocol.list('album','albumartistsort',row['base']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'album':
            self.protocol.find('album',row['base'],'albumartistsort',row['info']['albumartistsort']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'roottracks':
            self.protocol.list('artistsort').addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'artistsort':
            self.protocol.list('title','artistsort',row['base']).addCallback(self.reload_view).addErrback(self.handle_mpd_error)
        elif row['info']['type'] == 'playlist':
            pass
        elif row['info']['type'] == 'file':
            pass
        else:
            Logger.warn("Library: double click for ["+format(row)+"] not implemented")

    def handle_mpd_error(self,result):
        Logger.error('Library: MPDIdleHandler Callback error: {}'.format(result))

    def reload_row_after_sticker(self,copy_flag,index,result):
        self.rv.data[index]['copy_flag']=copy_flag
        self.rv.refresh_from_data()

    def set_copy_flag_find(self,copy_flag,index,result):
        for rrow in result:
            if copy_flag:
                Logger.debug("set_copy_flag_find: setting copy_flag to "+copy_flag+" for file "+rrow['file'])
                self.protocol.sticker_set('song',rrow['file'],'copy_flag',copy_flag).addCallback(partial(self.reload_row_after_sticker,copy_flag,index)).addErrback(self.handle_mpd_error)
            else:
                Logger.debug("set_copy_flag_find: clearing copy_flag for file "+rrow['file'])
                self.protocol.sticker_delete('song',rrow['file'],'copy_flag').addCallback(partial(self.reload_row_after_sticker,'',index)).addErrback(self.handle_mpd_error)

    def set_copy_flag_find_one(self,copy_flag,index,result):
        for rrow in result:
            if copy_flag:
                Logger.debug("set_copy_flag_find_one: setting copy_flag to "+copy_flag+" for file "+rrow['file'])
                self.protocol.sticker_set('song',rrow['file'],'copy_flag',copy_flag).addCallback(partial(self.reload_row_after_sticker,copy_flag,index)).addErrback(self.handle_mpd_error)
            else:
                Logger.debug("set_copy_flag_find_one: clearing copy_flag for file "+rrow['file'])
                self.protocol.sticker_delete('song',rrow['file'],'copy_flag').addCallback(partial(self.reload_row_after_sticker,'',index)).addErrback(self.handle_mpd_error)
            break

    def set_copy_flag(self,copy_flag):
        Logger.info('Library: set_copy_flag('+str(copy_flag)+')')
        for index in self.rbl.selected_nodes:
            row = self.rv.data[index]
            mtype=row['info']['type']
            Logger.info("Library: Setting copy_flag for "+mtype+" '"+row['base']+"' to "+copy_flag)
            if mtype == 'file':
                Logger.debug("set_copy_flag: adding uri or file")
                if copy_flag:
                    Logger.debug("set_copy_flag: setting copy_flag to "+copy_flag+" for file "+row['base'])
                    self.protocol.sticker_set('song',row['base'],'copy_flag',copy_flag).addCallback(partial(self.reload_row_after_sticker,copy_flag,index)).addErrback(self.handle_mpd_error)
                else:
                    Logger.debug("set_copy_flag: clearing copy_flag for file "+row['base'])
                    self.protocol.sticker_delete('song',row['base'],'copy_flag').addCallback(partial(self.reload_row_after_sticker,'',index)).addErrback(self.handle_mpd_error)
            elif mtype == 'albumartistsort':
                self.protocol.find(mtype,row['base']).addCallback(partial(self.set_copy_flag_find,copy_flag,index)).addErrback(self.handle_mpd_error)
            elif mtype == 'album':
                self.protocol.find(mtype,row['base'],'albumartistsort',row['info']['albumartistsort']).addCallback(partial(self.set_copy_flag_find,copy_flag,index)).addErrback(self.handle_mpd_error)
            elif mtype == 'artistsort':
                self.protocol.find(mtype,row['base']).addCallback(partial(self.set_copy_flag_find,copy_flag,index)).addErrback(self.handle_mpd_error)
            elif mtype == 'track':
                self.protocol.find('artistsort',row['info']['artistsort'],'title',row['base']).addCallback(partial(self.set_copy_flag_find_one,copy_flag,index)).addErrback(self.handle_mpd_error)
            elif mtype == 'playlist' or mtype == 'uri':
                Logger.info("Library: "+mtype+" copy_flag not implemented")
                #self.protocol.load(row['base'])
            else:
                Logger.warning("Library: "+mtype+' copy_flag not implemented')
        self.rbl.clear_selection()

    def rating_set(self,base,index,instance):
        Logger.debug('Application: rating_set('+instance.rating+')')
        instance.popup.dismiss()
        self.protocol.sticker_set('song',base,'rating',instance.rating).addCallback(partial(self.handle_rating_set,index,instance.rating,True)).addErrback(partial(self.handle_rating_set,index,instance.rating,False))

    def handle_rating_set(self,index,rating,succ,result):
        if succ:
            Logger.info("Library: successfully set song rating for "+self.rv.data[index]['base'])
            self.rv.data[index]['rating']=rating
            self.rv.refresh_from_data()
        else:
            Logger.info("Library: could not set song rating for "+self.rv.data[index]['base'])

    def generate_list(self,ltype):
        Logger.info(ltype+': generating with minimum stars '+self.ids.minimum_stars.text)
        if ltype=='rsync':
            self.rsync_data={}
            self.rsync_file=open(os.path.join(configdir,'rsync.inc'),'w')
        self.protocol.listallinfo('/').addCallback(partial(self.generate_list2,ltype)).addErrback(self.handle_mpd_error)

    def generate_list2(self,ltype,result):
        for row in result:
            if 'file' in row:
                uri=row['file']
                self.protocol.sticker_list('song',uri).addCallback(partial(self.list_add_uri,ltype,uri)).addErrback(partial(self.list_add_uri,ltype,uri))

    def list_add_uri(self,ltype,uri,result):
        docopy = False
        try:
            if 'rating' in result:
                if int(result['rating']) >= int(self.ids.minimum_stars.text):
                    docopy=True
            if 'copy_flag' in result:
                if ltype == 'rsync':
                    if result['copy_flag'] == 'Y':
                        docopy=True
                if result['copy_flag'] == 'N':
                    docopy=False
        except:
            docopy = False
        if docopy:
            self.ids.status.text=ltype+': '+uri
            Logger.debug(ltype+': '+uri)
            if ltype == 'rsync':
                wline=uri.encode("UTF-8")
                self.rsync_file.write(wline+"\n")
            elif ltype == 'playlist':
                self.protocol.playlistadd(self.ids.minimum_stars.text+" star or more",uri).addErrback(self.handle_mpd_error)

    def write_rsync(self):
        Logger.info('Rsync: writing to disk')
        self.rsync_file.close()

class LibraryRecycleBoxLayout(FocusBehavior,LayoutSelectionBehavior,RecycleBoxLayout):
    ''' Adds selection and focus behaviour to the view. '''

class LibraryRow(RecycleDataViewBehavior,BoxLayout):
    ''' Add selection support to the Label '''
    index = None
    selected = BooleanProperty(False)
    selectable = BooleanProperty(True)

    def rating_popup(self,instance):
        Logger.debug('Library: rating_popup()')
        layout = GridLayout(cols=2,spacing=10)
        popup = Popup(title='Rating',content=layout,size_hint=(0.8,1))
        for r in list(range(0,11)):
            btn=Button(font_name=resource_filename(__name__,os.path.join('resources','FontAwesome.ttf')))
            btn.text=App.get_running_app().songratings[str(r)]['stars']
            btn.rating=str(r)
            btn.popup=popup
            layout.add_widget(btn)
            btn.bind(on_press=partial(App.get_running_app().root.ids.library_tab.rating_set,instance.base,self.index))
            lbl=Label(text=App.get_running_app().songratings[str(r)]['meaning'],halign='left')
            layout.add_widget(lbl)
        popup.open()

    def refresh_view_attrs(self, rv, index, data):
        ''' Catch and handle the view changes '''
        self.index = index
        return super(LibraryRow, self).refresh_view_attrs(
            rv, index, data)

    def on_touch_down(self, touch):
        ''' Add selection on touch down '''
        if super(LibraryRow, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos) and self.selectable:
            # if we have a double-click, play from that location instead of selecting
            if touch.is_double_tap:
                Logger.debug("Library: double-click on "+str(self.index))
                App.get_running_app().root.ids.library_tab.rbl.clear_selection()
                App.get_running_app().root.ids.library_tab.handle_double_click(App.get_running_app().root.ids.library_tab.rv.data[self.index],self.index)
            else:
                return self.parent.select_with_touch(self.index, touch)

    def apply_selection(self, rv, index, is_selected):
        ''' Respond to the selection of items in the view. '''
        self.selected = is_selected
        lt=App.get_running_app().root.ids.library_tab
        if is_selected:
            lt.library_selection[index] = True
        else:
            if index in lt.library_selection:
                del lt.library_selection[index]

